package sim.app.species;

import sim.app.species.ANN.*;
import sim.engine.*;
import sim.field.continuous.*;
import sim.field.grid.*;
import sim.util.*;
import ec.util.*;
import java.util.*;
import java.io.*;

public class SpeciesSim extends SimState{

	public int fieldWidth, fieldHeight;
	protected boolean showLook;

	public Continuous2D creatures;
	public Continuous2D stationaryFood;
	public Continuous2D movingFood;
	public Continuous2D predators;
	public Continuous2D turrets;
	public Continuous2D bullets;

	private BestCreatureDisplayer bcd;
	
	private double discretization; // used to make the Continuous2D object

	private ConfigParams params;
	private PrintWriter statsFile;

	protected int numSpecies;

	private int runNum;
	
	public SpeciesSim(long seed, String paramFileName){

		super(new MersenneTwisterFast(seed), new Schedule(1));

		this.params = new ConfigParams(paramFileName);

		this.fieldWidth = this.params.getParamInt("FIELD_WIDTH");
		this.fieldHeight = this.params.getParamInt("FIELD_HEIGHT");
		this.showLook = this.params.getParamBoolean("SHOW_LOOK");
		this.discretization = this.params.getParamDouble("DISCRETIZATION");

		this.numSpecies = 0;

	}
	
	public SpeciesSim(long seed)
	{
		this(seed, "run.conf");
	}


	public void start(){

		super.start();

		this.openStatsFile();
		
		// add the schedulers that randomly add in new food
		this.schedule.scheduleRepeating(0, new StationaryFoodCreator(this.params.getParamInt("STATIONARY_FOOD_SPAWN_RATE"), 
									     this.params.getParamSpawnPointArray("STATIONARY_FOOD_SPAWN_POINTS")));
		this.schedule.scheduleRepeating(0, new MovingFoodCreator(this.params.getParamInt("MOVING_FOOD_SPAWN_RATE"), 
									 this.params.getParamSpawnPointArray("MOVING_FOOD_SPAWN_POINTS")));
		this.schedule.scheduleRepeating(0, new PredatorCreator(this.params.getParamInt("PREDATOR_SPAWN_RATE"), 
								       this.params.getParamSpawnPointArray("PREDATOR_SPAWN_POINTS")));

		this.bcd = new BestCreatureDisplayer(this.runNum);
		
		this.schedule.scheduleRepeating(0, this.bcd);

		this.creatures = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);
		this.stationaryFood = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);
		this.movingFood = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);
		this.predators = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);
		this.turrets = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);
		this.bullets = new Continuous2D(this.discretization, this.fieldWidth, this.fieldHeight);

		// maybe this should be done another way, for know, we'll leave it like this
		Axon.setMaxWeight(this.params.getParamDouble("AXON_MAX_WEIGHT"));
		Brain.setAxonWeightMutRate(this.params.getParamDouble("AXON_WEIGHT_MUT_RATE"));
		Brain.setNewSomaMutRate(this.params.getParamDouble("NEW_SOMA_MUT_RATE"));
		Brain.setNewAxonMutRate(this.params.getParamDouble("NEW_AXON_MUT_RATE"));
		Brain.setCompatibilityThreshold(this.params.getParamDouble("COMP_THRESHOLD"));

		Creature.setInputToBrainType(this.params.getParamString("CREATURE_BRAIN_INPUT_TYPE"));
		Creature.setCanSeeOtherCreatures(this.params.getParamBoolean("CREATURE_CAN_SEE_OTHER_CREATURES"));
		Creature.setVisionDivisions(this.params.getParamInt("CREATURE_VISION_DIVISIONS"));
		Creature.setLookMutRate(this.params.getParamDouble("CREATURE_LOOK_MUT_RATE"));
		Creature.setMatingEnergyThreshold(this.params.getParamInt("CREATURE_MATING_ENERGY_THRESHOLD"));
		Creature.setAsexualMatingThreshold(this.params.getParamInt("CREATURE_ASEXUAL_MATE_THRESHOLD"));
		Creature.setSharpAngleTurn(this.params.getParamDouble("CREATURE_SHARP_ANGLE_TURN"));
		Creature.setWideAngleTurn(this.params.getParamDouble("CREATURE_WIDE_ANGLE_TURN"));
		Creature.setSpeciesColors(this.params.getParamColorArray("CREATURE_SPECIES_COLORS"));
		Creature.setMatingRestriction(this.params.getParamString("CREATURE_MATING_RESTRICTION"));
		Creature.setSpeciesSimReference(this);

		StationaryFood.setEnergyAddedToCreature(this.params.getParamInt("STATIONARY_FOOD_ENERGY_ADDED_TO_CREATURE"));
		MovingFood.setRandomVectorLength(this.params.getParamDouble("MOVING_FOOD_RANDOM_VECTOR_LENGTH"));
		MovingFood.setEnergyAddedToCreature(this.params.getParamInt("MOVING_FOOD_ENERGY_ADDED_TO_CREATURE"));
		Predator.setRandomVectorLength(this.params.getParamDouble("PREDATOR_RANDOM_VECTOR_LENGTH"));
		Predator.setEnergyAddedToCreature(this.params.getParamInt("PREDATOR_ENERGY_ADDED_TO_CREATURE"));
		Turret.setShootInterval(this.params.getParamInt("TURRET_SHOOT_INTERVAL"));
		SpinningTurret.setShootInterval(this.params.getParamInt("SPINNING_TURRET_SHOOT_INTERVAL"));
		Bullet.setInitialEnergy(this.params.getParamInt("BULLET_INITIAL_ENERGY"));
		
		this.addToSim(this.creatures, "CREATURE");
		this.addToSim(this.stationaryFood, "STATIONARY_FOOD");
		this.addToSim(this.movingFood, "MOVING_FOOD");
		this.addToSim(this.predators, "PREDATOR");
		this.addToSim(this.turrets, "TURRET");
		this.addToSim(this.turrets, "SPINNING_TURRET");
		this.addToSim(this.bullets, "BULLET");

	}

	private void openStatsFile(){

		try{
			BufferedReader br = new BufferedReader(new FileReader("output/run.num"));
			this.runNum = Integer.parseInt(br.readLine());
			
		}catch(IOException e){
			this.runNum = 0;
		}

		try{
			PrintWriter pr = new PrintWriter(new BufferedWriter(new FileWriter("output/run.num")));
			pr.println(this.runNum+1);
			pr.close();
		}catch(IOException e){}

		try{
			this.statsFile = new PrintWriter(new BufferedWriter(new FileWriter("output/run" + this.runNum + ".out")));
			this.statsFile.println(this.statsFileHeader());
		}catch (IOException e){
			e.printStackTrace();
			System.exit(0);
		}

	}

	/* the functions above call this to actually add everything to the simulation
	 **/
	private void addToSim(Continuous2D cont2D, String entityType){

		int numEntities = this.params.getParamInt(entityType + "_NUM");
		SpawnPoint[] spawnPoints = this.params.getParamSpawnPointArray(entityType + "_SPAWN_POINTS");
		double senseDist = this.params.getParamDouble(entityType + "_SENSE_DISTANCE");
		double minDiam = this.params.getParamDouble(entityType + "_DIAMETER_MIN");
		double maxDiam = this.params.getParamDouble(entityType + "_DIAMETER_MAX");
		double speed = this.params.getParamDouble(entityType + "_SPEED");
		boolean[] look = this.params.getParamBooleanArray(entityType + "_LOOK");
		int lifeSpan = this.params.getParamInt(entityType + "_LIFE_SPAN");
		MutableDouble2D initPosition;
		MutableDouble2D initHeading;
		double diameter;

		if(entityType.equals("CREATURE")){
			Creature.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("STATIONARY_FOOD")){
			StationaryFood.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("MOVING_FOOD")){
			MovingFood.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("PREDATOR")){
			Predator.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("TURRET")){
			Turret.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("SPINNING_TURRET")){
			SpinningTurret.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else if(entityType.equals("BULLET")){
			Bullet.setDefaults(showLook, maxDiam, speed, look, senseDist, lifeSpan);
		}else{
			throw new RuntimeException("Something wrong happened in SpeciesSim when trying to add a new Entity of unknown type.");
		}

		// put the creatures in the simulation
		Entity newEntity;

		for(int i = 0; i < numEntities; i++){

			initPosition = spawnPoints[this.random.nextInt(spawnPoints.length)].getRandomLocation(this.random);

			/* obtain an initial heading
			 **/
			while((initHeading = new MutableDouble2D(this.random.nextDouble(), this.random.nextDouble())).length() > 1.0) ;

			if(this.random.nextBoolean()) initHeading.x *= -1.0;
			if(this.random.nextBoolean()) initHeading.y *= -1.0;

			if(initHeading.lengthSq() > 0.0) initHeading.normalize();

			/* obtain the diameter
			 **/
			diameter = this.random.nextDouble() * (maxDiam-minDiam);
			diameter += minDiam;

			/* create the creature and set all of its parameters
			 * There must be a better way of doing this, for now, this will work
			 **/
			if(entityType.equals("CREATURE")){

				int minEnergy = this.params.getParamInt("CREATURE_MIN_ENERGY_START");
				int maxEnergy = this.params.getParamInt("CREATURE_MAX_ENERGY_START");
				int energy = minEnergy + this.random.nextInt(maxEnergy-minEnergy);

				newEntity = new Creature(initPosition, initHeading, this.createBrain());
				newEntity.setAge(0);
				newEntity.setEnergy(energy);
				
			}else if(entityType.equals("STATIONARY_FOOD")){
				newEntity = new StationaryFood(initPosition, initHeading);
				newEntity.setAge(this.random.nextInt(newEntity.getLifeSpan()));
			}else if(entityType.equals("MOVING_FOOD")){
				newEntity = new MovingFood(initPosition, initHeading);
				newEntity.setAge(this.random.nextInt(newEntity.getLifeSpan()));
			}else if(entityType.equals("PREDATOR")){
				newEntity = new Predator(initPosition, initHeading);
				newEntity.setAge(this.random.nextInt(newEntity.getLifeSpan()));
			}else if(entityType.equals("SPINNING_TURRET")){
				newEntity = new SpinningTurret(initPosition, initHeading);
			}else if(entityType.equals("TURRET")){
				newEntity = new Turret(initPosition, initHeading);
			}else{
				throw new RuntimeException("Something wrong happened in SpeciesSim when trying to add a new Entity of unknown type.");
			}

			newEntity.setShowLook(showLook);
			newEntity.setDiameter(diameter);
			newEntity.setSpeed(speed);
			newEntity.setLook(look);
			newEntity.setSenseDistance(senseDist);

			cont2D.setObjectLocation(newEntity, new Double2D(initPosition));
			newEntity.stoppable = this.schedule.scheduleRepeating(0, newEntity);

		}

	}

	public void finish(){
		this.statsFile.close();
		this.bcd.closeFile();
	}

	/* stops the Entity's scheduler and removes it from the Continuous2D
	 **/
	public void removeEntityFromSim(Entity entity, Continuous2D cont2D){

		if(entity instanceof Creature){
			Creature c = (Creature)entity;

			statsFile.print(this.schedule.getSteps() + "\t");
			statsFile.print(c.getID() + "\t");
			statsFile.print(c.getAge() + "\t");
			statsFile.print(c.getStationaryFoodEaten() + "\t");
			statsFile.print(c.getMovingFoodEaten() + "\t");
			statsFile.print(c.brain.getSomaChromeLen() + "\t");
			statsFile.print(c.brain.getAxonChromeLen() + "\t");
			statsFile.print(this.creatures.getAllObjects().numObjs + "\t");
			statsFile.print(this.stationaryFood.getAllObjects().numObjs + "\t");
			statsFile.print(this.movingFood.getAllObjects().numObjs + "\t");
			statsFile.println(this.numSpecies);

		}

		entity.setAlive(false);
		entity.stoppable.stop();
		cont2D.remove(entity);

	}

	private String statsFileHeader(){

		String ret;

		ret = "Time\tID\tAge\tS_eat\tM_eat\tS_len\tA_len\tCLive\tSLive\tMLive\tSpcis";

		return ret;

	}

	/* creates a new, fully connected brain
	 **/
	private Brain createBrain(){

		String brainFileName = this.params.getParamString("CREATURE_BRAIN_FILE");
		int divisions = Creature.numVisionDivisions();
		int numInputs = Creature.numInputsToBrain();
		int numOutputs = Creature.numActions();
		String brainType = this.params.getParamString("CREATURE_BRAIN_TYPE");

		if(brainType.equalsIgnoreCase("AdamBrain")){

			if(brainFileName.equals("")) return AdamBrain.makeBrain(numInputs, numOutputs, this.random);
			else			    return AdamBrain.makeBrain(brainFileName, this.random);

		}else if(brainType.equalsIgnoreCase("DavidBrain") || brainType.equalsIgnoreCase("DaveBrain")){

			if(brainFileName.equals("")) return DavidBrain.makeBrain(numInputs, numOutputs, this.random);
			else			    return DavidBrain.makeBrain(brainFileName, this.random);

		}else{
			throw new RuntimeException(brainType + " brain not found.");
		}

	}

	public static void main(String[] args){
		doLoop(SpeciesSim.class, args);
		System.exit(0);
	}

	// setters and getters
	public int getNumCreatures(){

		if(this.creatures == null) return 0;
		else return this.creatures.getAllObjects().numObjs;

	}

	public int getNumSpecies(){
		return this.numSpecies;
	}

	public double getCompatibilityThreshold(){
		return Brain.getCompatibilityThreshold();
	}

	public boolean getShowLook(){ return this.showLook; }

	public void setShowLook(boolean b){

		this.updateShowLookOfBag(this.creatures.getAllObjects(), b);
		this.updateShowLookOfBag(this.stationaryFood.getAllObjects(), b);
		this.updateShowLookOfBag(this.movingFood.getAllObjects(), b);
		this.updateShowLookOfBag(this.predators.getAllObjects(), b);
		this.updateShowLookOfBag(this.turrets.getAllObjects(), b);
		this.updateShowLookOfBag(this.bullets.getAllObjects(), b);
		this.updateShowLookOfBag(this.creatures.getAllObjects(), b);

		this.showLook = b;

	}

	/* updates the showLook value of all entities in the bag
	 **/
	private void updateShowLookOfBag(Bag bag, boolean b){

		for(int i = 0; i < bag.numObjs; i++){
			((Entity)bag.objs[i]).setShowLook(b);
		}

	}

}
